# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

class PwpLicence:
    def __init__(self):
        pass

    @staticmethod
    def print():
        print("This program and library is licenced under the European Union Public Licence v1.2")
        print("")
        print("The full text of EUPL V1.2 can be found, in several languages at:")              # noqa
        print("https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12")
        print("")
        print("and was published on the Official Journal of the European Union 19.5.2017")
        print("")
        print("                      EUROPEAN UNION PUBLIC LICENCE v. 1.2")
        print("                      EUPL © the European Union 2007, 2016")                      # noqa
        print("")
        print("This European Union Public Licence (the ‘EUPL’) applies to the Work (as defined") # noqa
        print("below) which is provided under the terms of this Licence. Any use of the Work,")
        print("other than as authorised under this Licence is prohibited (to the extent such")
        print("use is covered by a right of the copyright holder of the Work).")
        print("")
        print("The Work is provided under the terms of this Licence when the Licensor (as")
        print("defined below) has placed the following notice immediately following the")
        print("copyright notice for the Work:")
        print("")
        print("        Licensed under the EUPL")                                                             # noqa
        print("")
        print("or has expressed by any other means his willingness to license under the EUPL.")             # noqa
        print("")
        print("1. Definitions")
        print("")
        print("In this Licence, the following terms have the following meaning:")
        print("")
        print("- ‘The Licence’: this Licence.")
        print("")
        print("- ‘The Original Work’: the work or software distributed or communicated by the")
        print("  Licensor under this Licence, available as Source Code and also as Executable")
        print("  Code as the case may be.")
        print("")
        print("- ‘Derivative Works’: the works or software that could be created by the")
        print("  Licensee, based upon the Original Work or modifications thereof. This Licence")
        print("  does not define the extent of modification or dependence on the Original Work")
        print("  required in order to classify a work as a Derivative Work; this extent is")
        print("  determined by copyright law applicable in the country mentioned in Article 15.")
        print("")
        print("- ‘The Work’: the Original Work or its Derivative Works.")
        print("")
        print("- ‘The Source Code’: the human-readable form of the Work which is the most")
        print("  convenient for people to study and modify.")
        print("")
        print("- ‘The Executable Code’: any code which has generally been compiled and which is")
        print("  meant to be interpreted by a computer as a program.")
        print("")
        print("- ‘The Licensor’: the natural or legal person that distributes or communicates")
        print("  the Work under the Licence.")
        print("")
        print("- ‘Contributor(s)’: any natural or legal person who modifies the Work under the")
        print("  Licence, or otherwise contributes to the creation of a Derivative Work.")
        print("")
        print("- ‘The Licensee’ or ‘You’: any natural or legal person who makes any usage of")
        print("  the Work under the terms of the Licence.")
        print("")
        print("- ‘Distribution’ or ‘Communication’: any act of selling, giving, lending,")
        print("  renting, distributing, communicating, transmitting, or otherwise making")
        print("  available, online or offline, copies of the Work or providing access to its")
        print("  essential functionalities at the disposal of any other natural or legal")
        print("  person.")
        print("")
        print("2. Scope of the rights granted by the Licence")
        print("")
        print("The Licensor hereby grants You a worldwide, royalty-free, non-exclusive,")
        print("sub-licensable licence to do the following, for the duration of copyright vested")
        print("in the Original Work:")
        print("")
        print("- use the Work in any circumstance and for all usage,")
        print("- reproduce the Work,")
        print("- modify the Work, and make Derivative Works based upon the Work,")
        print("- communicate to the public, including the right to make available or display")
        print("  the Work or copies thereof to the public and perform publicly, as the case may")
        print("  be, the Work,")
        print("- distribute the Work or copies thereof,")
        print("- lend and rent the Work or copies thereof,")
        print("- sublicense rights in the Work or copies thereof.")
        print("")
        print("Those rights can be exercised on any media, supports and formats, whether now")
        print("known or later invented, as far as the applicable law permits so.")
        print("")
        print("In the countries where moral rights apply, the Licensor waives his right to")
        print("exercise his moral right to the extent allowed by law in order to make effective")
        print("the licence of the economic rights here above listed.")
        print("")
        print("The Licensor grants to the Licensee royalty-free, non-exclusive usage rights to")
        print("any patents held by the Licensor, to the extent necessary to make use of the")
        print("rights granted on the Work under this Licence.")
        print("")
        print("3. Communication of the Source Code")
        print("")
        print("The Licensor may provide the Work either in its Source Code form, or as")
        print("Executable Code. If the Work is provided as Executable Code, the Licensor")
        print("provides in addition a machine-readable copy of the Source Code of the Work")
        print("along with each copy of the Work that the Licensor distributes or indicates, in")
        print("a notice following the copyright notice attached to the Work, a repository where")
        print("the Source Code is easily and freely accessible for as long as the Licensor")
        print("continues to distribute or communicate the Work.")
        print("")
        print("4. Limitations on copyright")
        print("")
        print("Nothing in this Licence is intended to deprive the Licensee of the benefits from")
        print("any exception or limitation to the exclusive rights of the rights owners in the")
        print("Work, of the exhaustion of those rights or of other applicable limitations")
        print("thereto.")
        print("")
        print("5. Obligations of the Licensee")
        print("")
        print("The grant of the rights mentioned above is subject to some restrictions and")
        print("obligations imposed on the Licensee. Those obligations are the following:")
        print("")
        print("Attribution right: The Licensee shall keep intact all copyright, patent or")
        print("trademarks notices and all notices that refer to the Licence and to the")
        print("disclaimer of warranties. The Licensee must include a copy of such notices and a")
        print("copy of the Licence with every copy of the Work he/she distributes or")
        print("communicates. The Licensee must cause any Derivative Work to carry prominent")
        print("notices stating that the Work has been modified and the date of modification.")
        print("")
        print("Copyleft clause: If the Licensee distributes or communicates copies of the")
        print("Original Works or Derivative Works, this Distribution or Communication will be")
        print("done under the terms of this Licence or of a later version of this Licence")
        print("unless the Original Work is expressly distributed only under this version of the")
        print("Licence — for example by communicating ‘EUPL v. 1.2 only’. The Licensee")                 # noqa
        print("(becoming Licensor) cannot offer or impose any additional terms or conditions on")
        print("the Work or Derivative Work that alter or restrict the terms of the Licence.")
        print("")
        print("Compatibility clause: If the Licensee Distributes or Communicates Derivative")
        print("Works or copies thereof based upon both the Work and another work licensed under")
        print("a Compatible Licence, this Distribution or Communication can be done under the")
        print("terms of this Compatible Licence. For the sake of this clause, ‘Compatible")
        print("Licence’ refers to the licences listed in the appendix attached to this Licence.")
        print("Should the Licensee's obligations under the Compatible Licence conflict with")
        print("his/her obligations under this Licence, the obligations of the Compatible")
        print("Licence shall prevail.")
        print("")
        print("Provision of Source Code: When distributing or communicating copies of the Work,")
        print("the Licensee will provide a machine-readable copy of the Source Code or indicate")
        print("a repository where this Source will be easily and freely available for as long")
        print("as the Licensee continues to distribute or communicate the Work.")
        print("")
        print("Legal Protection: This Licence does not grant permission to use the trade names,")
        print("trademarks, service marks, or names of the Licensor, except as required for")
        print("reasonable and customary use in describing the origin of the Work and")
        print("reproducing the content of the copyright notice.")
        print("")
        print("6. Chain of Authorship")
        print("")
        print("The original Licensor warrants that the copyright in the Original Work granted")
        print("hereunder is owned by him/her or licensed to him/her and that he/she has the")
        print("power and authority to grant the Licence.")
        print("")
        print("Each Contributor warrants that the copyright in the modifications he/she brings")
        print("to the Work are owned by him/her or licensed to him/her and that he/she has the")
        print("power and authority to grant the Licence.")
        print("")
        print("Each time You accept the Licence, the original Licensor and subsequent")
        print("Contributors grant You a licence to their contributions to the Work, under the")
        print("terms of this Licence.")
        print("")
        print("7. Disclaimer of Warranty")
        print("")
        print("The Work is a work in progress, which is continuously improved by numerous")
        print("Contributors. It is not a finished work and may therefore contain defects or")
        print("‘bugs’ inherent to this type of development.")
        print("")
        print("For the above reason, the Work is provided under the Licence on an ‘as is’ basis")
        print("and without warranties of any kind concerning the Work, including without")
        print("limitation merchantability, fitness for a particular purpose, absence of defects")
        print("or errors, accuracy, non-infringement of intellectual property rights other than")
        print("copyright as stated in Article 6 of this Licence.")
        print("")
        print("This disclaimer of warranty is an essential part of the Licence and a condition")
        print("for the grant of any rights to the Work.")
        print("")
        print("8. Disclaimer of Liability")
        print("")
        print("Except in the cases of wilful misconduct or damages directly caused to natural")
        print("persons, the Licensor will in no event be liable for any direct or indirect,")
        print("material or moral, damages of any kind, arising out of the Licence or of the use")
        print("of the Work, including without limitation, damages for loss of goodwill, work")
        print("stoppage, computer failure or malfunction, loss of data or any commercial")
        print("damage, even if the Licensor has been advised of the possibility of such damage.")
        print("However, the Licensor will be liable under statutory product liability laws as")
        print("far such laws apply to the Work.")
        print("")
        print("9. Additional agreements")
        print("")
        print("While distributing the Work, You may choose to conclude an additional agreement,")
        print("defining obligations or services consistent with this Licence. However, if")
        print("accepting obligations, You may act only on your own behalf and on your sole")
        print("responsibility, not on behalf of the original Licensor or any other Contributor,")
        print("and only if You agree to indemnify, defend, and hold each Contributor harmless")
        print("for any liability incurred by, or claims asserted against such Contributor by")
        print("the fact You have accepted any warranty or additional liability.")
        print("")
        print("10. Acceptance of the Licence")
        print("")
        print("The provisions of this Licence can be accepted by clicking on an icon ‘I agree’")
        print("placed under the bottom of a window displaying the text of this Licence or by")
        print("affirming consent in any other similar way, in accordance with the rules of")
        print("applicable law. Clicking on that icon indicates your clear and irrevocable")
        print("acceptance of this Licence and all of its terms and conditions.")
        print("")
        print("Similarly, you irrevocably accept this Licence and all of its terms and")
        print("conditions by exercising any rights granted to You by Article 2 of this Licence,")
        print("such as the use of the Work, the creation by You of a Derivative Work or the")
        print("Distribution or Communication by You of the Work or copies thereof.")
        print("")
        print("11. Information to the public")
        print("")
        print("In case of any Distribution or Communication of the Work by means of electronic")
        print("communication by You (for example, by offering to download the Work from a")
        print("remote location) the distribution channel or media (for example, a website) must")
        print("at least provide to the public the information requested by the applicable law")
        print("regarding the Licensor, the Licence and the way it may be accessible, concluded,")
        print("stored and reproduced by the Licensee.")
        print("")
        print("12. Termination of the Licence")
        print("")
        print("The Licence and the rights granted hereunder will terminate automatically upon")
        print("any breach by the Licensee of the terms of the Licence.")
        print("")
        print("Such a termination will not terminate the licences of any person who has")
        print("received the Work from the Licensee under the Licence, provided such persons")
        print("remain in full compliance with the Licence.")
        print("")
        print("13. Miscellaneous")
        print("")
        print("Without prejudice of Article 9 above, the Licence represents the complete")
        print("agreement between the Parties as to the Work.")
        print("")
        print("If any provision of the Licence is invalid or unenforceable under applicable")
        print("law, this will not affect the validity or enforceability of the Licence as a")
        print("whole. Such provision will be construed or reformed so as necessary to make it")
        print("valid and enforceable.")
        print("")
        print("The European Commission may publish other linguistic versions or new versions of")
        print("this Licence or updated versions of the Appendix, so far this is required and")
        print("reasonable, without reducing the scope of the rights granted by the Licence. New")
        print("versions of the Licence will be published with a unique version number.")
        print("")
        print("All linguistic versions of this Licence, approved by the European Commission,")
        print("have identical value. Parties can take advantage of the linguistic version of")
        print("their choice.")
        print("")
        print("14. Jurisdiction")
        print("")
        print("Without prejudice to specific agreement between parties,")
        print("")
        print("- any litigation resulting from the interpretation of this License, arising")
        print("  between the European Union institutions, bodies, offices or agencies, as a")
        print("  Licensor, and any Licensee, will be subject to the jurisdiction of the Court")
        print("  of Justice of the European Union, as laid down in article 272 of the Treaty on")
        print("  the Functioning of the European Union,")
        print("")
        print("- any litigation arising between other parties and resulting from the")
        print("  interpretation of this License, will be subject to the exclusive jurisdiction")
        print("  of the competent court where the Licensor resides or conducts its primary")
        print("  business.")
        print("")
        print("15. Applicable Law")
        print("")
        print("Without prejudice to specific agreement between parties,")
        print("")
        print("- this Licence shall be governed by the law of the European Union Member State")
        print("  where the Licensor has his seat, resides or has his registered office,")
        print("")
        print("- this licence shall be governed by Belgian law if the Licensor has no seat,")
        print("  residence or registered office inside a European Union Member State.")
        print("")
        print("Appendix")
        print("")
        print("‘Compatible Licences’ according to Article 5 EUPL are:")                      # noqa
        print("")
        print("- GNU General Public License (GPL) v. 2, v. 3")
        print("- GNU Affero General Public License (AGPL) v. 3")                             # noqa
        print("- Open Software License (OSL) v. 2.1, v. 3.0")
        print("- Eclipse Public License (EPL) v. 1.0")
        print("- CeCILL v. 2.0, v. 2.1")                                                         # noqa
        print("- Mozilla Public Licence (MPL) v. 2")
        print("- GNU Lesser General Public Licence (LGPL) v. 2.1, v. 3")
        print("- Creative Commons Attribution-ShareAlike v. 3.0 Unported (CC BY-SA 3.0) for")               # noqa
        print("  works other than software")
        print("- European Union Public Licence (EUPL) v. 1.1, v. 1.2")                               # noqa
        print("- Québec Free and Open-Source Licence — Reciprocity (LiLiQ-R) or Strong")                         # noqa
        print("  Reciprocity (LiLiQ-R+).")
        print("")
        print("The European Commission may update this Appendix to later versions of the above")
        print("licences without producing a new version of the EUPL, as long as they provide")                   # noqa
        print("the rights granted in Article 2 of this Licence and protect the covered Source")
        print("Code from exclusive appropriation.")
        print("")
        print("All other changes or additions to this Appendix require the production of a new")
        print("EUPL version.")                                                                             # noqa
    # End of patched text
