# file generated by pwpImageEmbedder from pwpLogo.png

import io
from piwiPre.pwpImageEmbedder import PwpBioImage

pwpLogo_png = PwpBioImage(io.BytesIO(b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x002\x00\x00\x00\x1c\x08\x06\x00'	# noqa
                                     b'\x00\x00#\x99\xec\xf7\x00\x00\n)IDATx\x9c\xc5\x98{\x8c\\\xd5y\xc0\x7f\xdf9g\xee'	# noqa
                                     b'<w\xd6\xbb\xeb56\xebW\x9d`(\xc6`\x92\xb8\xb4\x81(\xde\x84\x06\x1a\xd2\x02Fr\xd2'	# noqa
                                     b'$U\xd2\xa8U\xab>B\xa3\x8a\xb6\x12\xaal\xa7\x02"\x84\x14\xa9y\x82\xd2\xd0T\x91'	# noqa
                                     b'\xdd\xae\x83B\xd2\x00\r\x01\x0c1uE\xc3\xcb\xa1k\x1c\x83\xe3]\xbc~\xaew\xd7\x9e'	# noqa
                                     b'\x9d\xc7\xbdw\xee9_\xff\x98]?\xc0&`\\\xf5\x93\x8e\xe6j4\xf7;\xdf\xef{\x9d\xef'	# noqa
                                     b'\x8c\xf0\xff$\xfa\xb7|\x99\x88\xbfz%\xbb(\xa3\xd8\xe3\x04\x0f\x08\xa0\x04,\xaeq'	# noqa
                                     b'([\x96\x7f\xcd\xf9\xd8\xdc\xe6\xee\x0e\xf7\xfc*}\xf2\x8e\r\x02\x11P\xfd;\x96'	# noqa
                                     b'\x93c\x01\x19\x810\xa3\xd7\x15:kV\xb2\xb8\xb3\xc0`\xcdw\x92zX\xfc\xd2\xdaG\xc2'	# noqa
                                     b'\xca\x0f^o\xf4\x04F\xe7\xf3\x85\xef}5\\\xb5\xfd/\x8d\x14\xdd\x0e|\xf6y\xc0\xa0'	# noqa
                                     b"\xa4\xb4xN\xbeB\xf2z;\xdcy\x81XOD\xcc\xa3X\x96\x10\x00'\x1ds\x92\x18\x1a\xf1I"	# noqa
                                     b'\xeb\xf2@\xde\x82\x06\xf0J\x10\x08{\xb7\x99CE\x8f\x04\x0f\xc6\x80\x06T\x1c\xf6'	# noqa
                                     b'\xd0\x8b&(X\xd1+\xc8\xf1\x14\x01\xe8\xf8\xe4\xe3\xc0\x90\xae\xc7\xc9F\xb2\xf3'	# noqa
                                     b'\x02r\x8a\x18\xc0\x92\x10\x08\x12\xf0j|\xa2\xd8\xe5\xab`\xf9\x07\xa1\xd0\x0b'	# noqa
                                     b"\x8d\xc3\xb0\xf3'\xf8\x91W\xb0%@\x90b\x11Y\xb5\xe3N\xda\xcf\x9d\x9e\x1a\n\x0cD`"	# noqa
                                     b'\x0b@\xdb\xeb\xccW\x19\x01\x87P:\x93\x01\xe7\x0b\x04\x84\x80\x11C\xaahq\x8e\xb1'	# noqa
                                     b'\x9f\xfb\x16\xfbJ\xef\xe5\x85\xdd\xafQ\x9fhP.\x15\xb9j\xdd_3\xff\xf0\xa3\x84'	# noqa
                                     b"\x7f\xfb\x0b\x8c\t\x00D\x05\x88\xe4\x0c\x19\xae\xdaYB'\xbc\x8aA1t\x8a\xe9\xff"	# noqa
                                     b"\x10DE\x08B\xc8\x950\xb7n\xe3\x81g'xx\xcb\x17\xe9\xe9\xaeR\xa9T\xe8\xe9\xe9\xe5"	# noqa
                                     b"?\x1f\x7f\x98\x0f\xfd\xde'\xb8\xfe\xcf\xfe\x9d\xf0\x8f\x1f\xc5\x14:\xa9\x84\xea"	# noqa
                                     b'\xdb\xd8\xe7\xccum\xde\xf4\x1dUQ\x1d\xb2\xbau\xbdS\x1d\xb2\xaa\xeb\xcf\xfe{c'	# noqa
                                     b"\xa1\x110k\xef\xe0\xb1=\xf0\xf5\xbbn'\x8e\x13F\xf7\xedg\xc7K\xc3\xfc\xfc\x7f"	# noqa
                                     b'\x86Y\xbd\xfa}<\xf6\xc0w\xf8Y\xbc\x02s\xed\x9f\x12\xea\x01\xcc\xf9\xf1\xe5\x1b'	# noqa
                                     b'\xb4\xa8\xae7<\x89\xd90H\x10\x91\x00o\x08\xa5\xe8\xd6\xf5\x965\x04d\xe3\x8c+'	# noqa
                                     b'\x97@6\n\xdd\x05\xb2\x8bn\xe6{\xb7\xdfCT,S\xaf\xd7QU\x8c1\x8c\x8d\x8d\xb1i\xd3f'	# noqa
                                     b'\xd6\xdet#?y\xe8A\xde\xf7\xc9\xcf\xc2c\xdf\xecD\xe4|\x80\xa8"l\x192\xf4\x0f\x8b'	# noqa
                                     b'\x0cn\xccD6\x06 t\xa0\xfe\xeb\x02\x8e\x8c\xae\xf2\xd3a\xbe\xadF\xe3\xf4\xf7\x0c'	# noqa
                                     b'\x8b\xfb\xf0\xa8\x0cn\xcc\x00t\xfdz\xa7\x1b6\x04\xfey\ra\xc7(f`\x11\x13M\xc3'	# noqa
                                     b'\xf8\xe1\x83DQ\x1e\x99\xc9}U\xa5R\xa9P\xab\xd5\xd8;2\x8a\xc9\x97h\xf8\n\x95j'	# noqa
                                     b'\x11m\xb5\x103\xdbx\xdf\x01\x88\x08\n\xeb<\x80DBH~\xb4\xb2\xbdk\xe2\xba\xf8`r'	# noqa
                                     b'\xdd\x91-\xcf\xae\xce\xb5\xa4\xdbz\xf0\x06\xda\xd1\xc1\xd6\xb1G\xee\xfby\xe5'	# noqa
                                     b'\xc2\xc2#v\xc5\xdc\xef\x8a|t\x0f\x1b7\xa2\xf7~\xcc`Eh\xb7\x88\x9c`\x9d\xa3\x9d'	# noqa
                                     b'\xc4\xa7\xe5~\x08\x01\xe7\x1c\xd3\xd35\xca6O\xce\x02Y\xc6\x99\xea\xfc\x9c@T\xd5'	# noqa
                                     b'\xd0\xfe\xe1o\xa4\xbbj7\xb4\x0e6\x7f\xe7\xf0w_}O1\xb1\x12\x1fOi6\x132\r\xc1\x18'	# noqa
                                     b'\t\x1aT\x8cH1\x8c%W\xc9Hz\x95\xdf9}[\xfa\xdf\xdf~0\xb7z\xdeWD>\xf6\x8c\xff\x02'	# noqa
                                     b'\x9e\xa3\x07\xe91S\xfc\xfaeW\xb2\xed\xf1\x87\xe8\xed\xeb\xa7\xddn\x03\x9d\xa8'	# noqa
                                     b'\x94+\x15\xc6\x0f\x1f\xe4\xe2U\xbfE\xbe\xb9\x87Pkc\xba,\x8436\xa2\xb7\x07r\xf4'	# noqa
                                     b'\xfb_{\xd64\xb9\xd2\xd6\x95f-\xa6\x19\xa74\x8cdb\x11\xcd\x89\xb1\x18\x83b\xc4\t'	# noqa
                                     b'\x806}[\x1bG\xd2 \x87(\xeb\xe1\xf6\xa7\xf4\x95\xc6\xa7\xea\xdb\xff\xe9A\xf3\xcc'	# noqa
                                     b'?\x14\xd9=\x02\x8f\xdf!\x9f\xfb\xf3\xfby\xf9\xc5g\x98\x988J\xb9\xab\x0b\x11\xc1'	# noqa
                                     b"g\x9e\xfa\xf1)\xe6]z9\xb7\xac\xfd]\xf4\xfe\x0f \x8ew\x9aQ'Aj;k\x03\x16\xa3\x9e"	# noqa
                                     b'\x90a\xc5\x9a\xc8\x08\xdai\x02r\xea&\x9dg\x11\x10\xb1bp\xe8T\xad\x19\x98\x0c'	# noqa
                                     b'\xb6:\xeeoJ\x16~\x819s\xff\x03\xf3\xf4\xbf\xca\x92\x0b.\xe7\x9eo|\x9bo\xddw//'	# noqa
                                     b'\xefx\x9e$\x89)U\xab|\xe0\xb7o\xe0\x8f>\xfbiz\x1e\xfd<\xfa\x8b\x9f!\x15s^\xa2'	# noqa
                                     b'\x01\xe0\xc4\xca\x13\xc5B\xf4\x89F\x9c\x18:\xf3\xcc[\x13E\x8c\x11K\xde\xeat\xb3'	# noqa
                                     b'\x11\xc2\xb0\x98\xc6\xc2\x1be\xde\x87\x96\x90\xff\xf1\xed,\xde\xbb\x8d/~\xf2o'	# noqa
                                     b'\x98\xf8\xcc-\xd4SeNA\xb4;\x1dQ6\xdf@\x18\xde\x86\xa9X\xf0o\x0b"\xa0\x04\xc2'	# noqa
                                     b"\x99-t\xaa\xfa\xaah'\x87\xe5\\*O\x11\x11c]\x01\xd2\xfd\xe3\x1c\xac_\xca\x9c\xd5"	# noqa
                                     b'wP\x9a|\x8ah\xf3-\xf4\xb9\x88>\x89\xa0y\\\xa8\x1f\x17,\x98>\x07Y\x06\xf6\x14=Bg'	# noqa
                                     b'\xd6\xeaLo\x10\xc2\xc9\x19\xcd\x18h\x87\x08\x0b\xde\xb9\xd2\xd6\xad\xb7;FF\x9c'	# noqa
                                     b'\xea\xd20\xd3eqb\xccKY\x08\x9c\x1b\xc5)<\n&\xef\xd0F\x8d\x89\x9d9\x8e\xcd\xb9'	# noqa
                                     b'\x89\xd2\xd2\xeb)0J)\x1e\xc6\xf9\x17j$\xbcJL\x12\x92\xacd\xf2\xa6x\xa2\xad\x89'	# noqa
                                     b'\x98\x10B \t\xd3FI\x01K\x9e*FLP\xc5$a\x8a2E\n\xf9\xc8\xeeK\x9e\x1e\xec\xb4\xff'	# noqa
                                     b'\x8c?\x04\xd5!\x0b\xeb\x82\xfc\xf2\xae\xbb\xaeP\xd5\xe7\x8c1V\xdf\xce\xa8\xf0&"'	# noqa
                                     b'\xa2h\x16\x08\xc1\xa0\xae@\xae\xbb\xa0\xa5~\xea\x85\xfeh{iq\xf1Gv\xd9\xdc\x87D'	# noqa
                                     b'\xae\xdb{\xfaK\xa0\xc7\x87\xae\x89\xc7\xb9\xbc\x90o\xefc\xe0\xc2GD\x06;\xe7\xd5'	# noqa
                                     b'\xd8\xa6\xcf4\x7fQ\xff\xe3VM\xe7\x15\xdfU~\xb0\xd4g_lO\xb38w\xa1\xfd\xa1T\xd7'	# noqa
                                     b'\xed\x04\x90\xc9/}\xa9{*\x84]\x91s\xf3\xd3,Sy\x87\x91y\xbdq\xa2J\xf0\x01QC\xb9T'	# noqa
                                     b'\xa0\xd0\x1d\x11\xe6\x98z\xe5\xdd\x95\x7fqW,\xbb\x83\xe6\x01\xf1\x07\xfc\xcd'	# noqa
                                     b'\xcd\x03\xcd\xb5\xed}\xcdA\x9dVz\xe6\x96i_ln\xcd\xaf\\\xb9\xc9\xbf8\xfc\xcdd'	# noqa
                                     b'\xb8y\xcb\xd4\xc84\xf9\xf9yz\xdeU\xa5\xb1\xaf\x89\xb4@\x17\xd9\xe9\xea\xd5\xfd'	# noqa
                                     b'\xb7R}\xcff\x01\xd8s\xe7\x9d\xcf\x97\xf3\xf9+\x1bi\x1a\xe4W\xcc_g\x15\xe5\xf47_'	# noqa
                                     b'\x1f\\\x01U\r\x9a\xa9\x1a\x15\xdb\xd7\xdfE{\xbeL\x15\x07J\x86\xddiw\xed\xb5i'	# noqa
                                     b"\xa6\xe3\x18W\xb4\xed\xf2\xdc\x92\xad\xac\xea\xda\xdf\x18m\x8a\x1b\xf3\x0b'\x8e"	# noqa
                                     b'\xd5}\xf7\x8anQU\x99\xde]\x0f\xa8\x12\xda\xaa=\x97v\xbb\xee\\\x91\xb8/<\xde\x99'	# noqa
                                     b'\xb5D\xc6\xad1\xb3\x15\x7fN\x1c\xe2\x84\xd0\x0e\xa8\x9fi\x1a\x86\xce\xe81\xab'	# noqa
                                     b"\xae3\x91w\xce#A\xc7'j\xc1\x1c\n=\xf1H\x9dha1caN\xba\xa3\xbc59\x93Kk)S?\x9d\\"	# noqa
                                     b'\x84B:\xc7\xf8\xdee\xbd6>\x12\xd3\xda\xdf\xc2D\xc6\xaaB\xbe\x9a\x83\x82\t{\xb7'	# noqa
                                     b'\x1f\xc8\x06\xae\xbe\xe0\xc3n\xc6Y\xe3F\x04\x119\xb7"Q(\x0e\x14\xb1\x05\x8boy'	# noqa
                                     b'\xb2FFV\xcf\xf0\xb1?3\x98"\xc6\x1a\xdb\xb5\xfa\xbd\x8a\xc9\xd1\xdc\xf5\xb2\xab.'	# noqa
                                     b'7\xc4\x87R\x1a#\r\xa2\x9e\x88\xdc@^Md4\xabe\xb6\xb6\xab\x86\x8f=&\xdf\tyH5\x14'	# noqa
                                     b'\x07\x8a!\xd9\xdbt\x8bW\xcc\x8b\xc2\x02\xb7\xc9\x01\xa8\xea\x01AB\xf0\x9a\xc9'	# noqa
                                     b"\x0c\xd9\x8c'e\xa6\x9b\x89\x08r\xd63F\xa01\xda\xc0F\x16W\xb6\xd8\x8a#\xea\x8d"	# noqa
                                     b'\x10\x01\x9f\x04\xb2\xfa)`\x01\xb4\xdd\xa6\xb8d\x11\xe5\x8b\x96\x0b\x1a(,\x98'	# noqa
                                     b'\xcf\xe4\x13OP\xbd(\x87\xeb\xca\x91M\xb7\x89\x0f\xc5\x92\xd53\xd1\xa0\x18g\xb0y'	# noqa
                                     b'K\x08\x1a\xb2f&\x0b.\x9fkJ\xdd\x05\xd3\x10w wY\xf9kv\xf9\xa7\xefv\x00\xd6\xb9'	# noqa
                                     b'\x1f\xd8\x82\xb9\xad\xbfR-\xa0@\x80\x90\x05|\xdb\xd3N=I\x9a\x91\x05\x1f\xc4J'	# noqa
                                     b'\x10#\x82`\xe0t0\x11!\xa4\x81\xa4\xe5\xd1\xf1\x041\x82\xc9\x1b\\\xc5\xe1*\x8e'	# noqa
                                     b'\xa8/B\x10B[I\xa7\x1a\x14\x17/\x00\x0cY\xbd\x81\xab\x96\xc8/\\L2\xfe\n\xc9\x94'	# noqa
                                     b'\x92N&\x98\xc8 N0\x18\x02\xaa>\xc9B9_\xb0]\xab{)\xae\xec\xdaN#l\xaa^\xbbz\xb3'	# noqa
                                     b'\xc8e\x93\xf0\x07\xc8\xcc9\xa8\x93O\xdfwc\xa9\xb7\xb4\x8c\xb6\xd7,\xf5\x15m\xeb'	# noqa
                                     b'\x02\xdf\xca.\xf4IX\xe4\xeb~\xa9\x8d\xb5\xcf\xb6 \x9eNi\xb6\x12\x82\xaa\x17gT'	# noqa
                                     b'\xcc\t\xa8\xd9\x18\x9e\x94\x00\xea\xb5Sz\xb3`\xe5\x1c\xae$D\xfd}D\x0b\xde\x8d'	# noqa
                                     b'\xed\xeaAr%&\x9fz\x9a\xa88Nc\xacMH\xfc\xac\x1e\r\x99z\x87q}\x03]\xd8\xa5\xf9'	# noqa
                                     b'\xe7\x8b\xd7\xf4\xfd\xbd\xe4n~\xf8DV\x0f\rY\xf3\xf1u~v\xdb\xb3_\x08,h\xf6L\x1f'	# noqa
                                     b'\xc7F\xaf\x88G\x9ak\xe2\xa3\xe9`{"\xb92\x8aM9\xd43\x9a\xf5\x84$m\xa3\x827\xd6('	# noqa
                                     b'\x1d\xb0\x93\xa3\xce\xeb\xc1\x82v\x8eA\xf5\xd8\xbc\x90\x9f\xdf\xab\x85\x85\xf3'	# noqa
                                     b'\t\xf18\xe9xS\x9bc-\xc4\xa2\xeaQ\xab\xe2z\xe7u\xc1\x02w\xa8\xba\xb2z7K\x7f\xff'	# noqa
                                     b'\xab"\xd2VT\xd8\xba\xc1\xb2f\x83\x9f\xad\xeb\x13\xdb\xe8\xd0\x90\xa5\xbf\xff'	# noqa
                                     b'\x94m\x9f\x84\xaf\xafP\xd9\xb2.\x9c\x06)\xa0\xe1\xa9E~\xdf\xfekZ\xfb[\xd7\xa6'	# noqa
                                     b'\x13\xe9\xfb\xd3\xa9\xf4\xe2r\xe6$\xabg\xb4\x1a\t\xad\xb4\xad\xe2D\xc5\x88\x01P'	# noqa
                                     b'\xa5\xf3OH\xe8<\xe9L\x04E\x112\x8f1\x90+\xe7\xd1Tq9\x8b\xcbY\n\xd5\x08\xdfk\xa6'	# noqa
                                     b'\xba/\xe9\xba\x97K~\xed\xcb"W\x1f\xe9\xe8\x1a\xb2"\xeb\xde0\xa4\xbd\xa5^\xab'	# noqa
                                     b'\xaa\xc2\x96-\x86\xfeaap\xa3\x97S\xc0T\xd5\xc0C\x17\xa7{\x8e\xfff6\x9e^\x13\x1f'	# noqa
                                     b'm\xbd\xdfO\xb4/\tS\x9ef=\x06\xc09K.r\xe4\xf2\x16\x89\x0cj\xc1\x8b\xe2\x8d\x06'	# noqa
                                     b'\xe3$\x11kZ\x88\xb6Ld[\x923\xd3\xb9.7Z]P~\x82\xe5s\x1f\x10\xf9\xc8\x81Y\x00X'	# noqa
                                     b"\x17\xe4,-\xe7-\xdd\xfcg\xc2w\xc2\x0b\xb3\xf7\xfa'\x9f\x04\x11\xc9\x80\x97g\xd6"	# noqa
                                     b'\xfd\xaa\x9a\xa7\xf6\xc0G\xe2\xe1\xda\x9f\xb8C\xadA\x9f\x05\xeb\xca\xee\x80-'	# noqa
                                     b'\xd9_\xba\xa2\xdb\xed\xaa\xb9W]\xd9\x8e\xd9\x92;\xca\xdc\xc21p\xc7!\xaa\xc3@'	# noqa
                                     b"\x13.IL\xde\xa4\x9a\x9e\xb4u\xeb\xd6\xf5nM'\x85\xdetT\xfe_P,\xf7:\x15\x95_I\x00"	# noqa
                                     b'\x00\x00\x00IEND\xaeB`\x82'),	# noqa
               img_format="png",	# noqa
               filename="pwpLogo.png")	# noqa
